# -*- coding: utf-8 -*-
"""
Calculate agreement for coding of competencies and means between first and second coder.

Required files:
    - user_responses_for_coding_final_labels_and_explanations_only_anonym_double_coding.xlsx
    - user_responses_for_coding_final_labels_and_explanations_only_anonym_pa_coded.xlsx
    - user_responses_for_coding_final_labels_and_explanations_only_anonym_smoking_coded.xlsx
    
Created files:
    - calculate_agreement_between_coders_output.txt

Author: Nele Albers
Date: September 2023
"""


import pandas as pd
from sklearn.metrics import cohen_kappa_score


def brennan_prediger(codes1, codes2, q):
    """
    Calculates the Brennan-Prediger coefficient.

    Parameters
    ----------
    codes1 : list
        Codes given by first coder with possible codes 0, ...., q-1.
    codes2 : list
        Codes given by second coder with possible codes 0, ..., q-1.
    q : int
        Number of possible codes/response categories. Cannot be 1.

    Returns
    -------
    float
        Brennan-Prediger coefficient.

    """

    p_a = sum([1 for i in range(len(codes1)) if codes1[i] == codes2[i]])/len(codes1)
    p_e = 1/q  # q is the number of response categories

    return (p_a - p_e)/(1 - p_e)


if __name__ == "__main__":

    
    # Load data from the two coders
    df_coder_2_s = pd.read_excel("user_responses_for_coding_final_labels_and_explanations_only_anonym_double_coding.xlsx",
                                 sheet_name = "Final_Coding_Smoking")
    df_coder_2_pa = pd.read_excel("user_responses_for_coding_final_labels_and_explanations_only_anonym_double_coding.xlsx",
                                 sheet_name = "Final_Coding_PA")
    num_responses_double_coded = len(df_coder_2_s) + len(df_coder_2_pa)

    df_coder_1_s = pd.read_excel("user_responses_for_coding_final_labels_and_explanations_only_anonym_smoking_coded.xlsx")
    # Only take those responses coded by the first coder that the second coder also coded
    df_coder_1_s = df_coder_1_s[df_coder_1_s["Explanation"].isin(df_coder_2_s["Explanation"].to_list())]
    df_coder_1_pa = pd.read_excel("user_responses_for_coding_final_labels_and_explanations_only_anonym_pa_coded.xlsx")
    # Only take those responses coded by the first coder that the second coder also coded
    df_coder_1_pa = df_coder_1_pa[df_coder_1_pa["Explanation"].isin(df_coder_2_pa["Explanation"].to_list())]


    # 0 = competency, 1 = means, 2 = unclear/mixed
    df_coder_2_s["Labels"] = [0 if df_coder_2_s.iloc[i]["Competency"] == 1 else 1 if df_coder_2_s.iloc[i]["Means"] == 1 else 2 for i in range(len(df_coder_2_s))]
    df_coder_2_pa["Labels"] = [0 if df_coder_2_pa.iloc[i]["Competency"] == 1 else 1 if df_coder_2_pa.iloc[i]["Means"] == 1 else 2 for i in range(len(df_coder_2_pa))]
    df_coder_1_s["Labels"] = [0 if df_coder_1_s.iloc[i]["Competency"] == 1 else 1 if df_coder_1_s.iloc[i]["Means"] == 1 else 2 for i in range(len(df_coder_1_s))]
    df_coder_1_pa["Labels"] = [0 if df_coder_1_pa.iloc[i]["Competency"] == 1 else 1 if df_coder_1_pa.iloc[i]["Means"] == 1 else 2 for i in range(len(df_coder_1_pa))]


    # Compute Cohen's kappa based on the 3 possible labels
    cohens_kappa_result = cohen_kappa_score(df_coder_2_pa["Labels"].to_list() + df_coder_2_s["Labels"].to_list(), 
                                            df_coder_1_pa["Labels"].to_list() + df_coder_1_s["Labels"].to_list())
    print("Cohen's kappa:", round(cohens_kappa_result, 2))


    # Compute Brennan Prediger Kappa for labels 0-2 (i.e., the 3 possible labels)
    num_cat = 3  # number of possible labels/categories
    brennan_prediger_result = brennan_prediger(df_coder_2_pa["Labels"].to_list() + df_coder_2_s["Labels"].to_list(), 
                                               df_coder_1_pa["Labels"].to_list() + df_coder_1_s["Labels"].to_list(), num_cat)
    print("Brennan-Prediger kappa:", round(brennan_prediger_result, 2))

    
    # Write to file
    with open("calculate_agreement_between_coders_output.txt", 'w') as f:
        f.write("Number of double-coded responses: " + str(num_responses_double_coded) + "\n")
        f.write("Cohen's kappa: " + str(round(cohens_kappa_result, 2)) + "\n")
        f.write("Brennan-Prediger kappa: " + str(round(brennan_prediger_result, 2)) + "\n")
