# -*- coding: utf-8 -*-
"""
Process data from expert questionnaire for the factor analysis.

Required files: 
    - Expert_Competencies_Final_Qualtrics_Export_2022_12_22_anonym.xlsx

Output files:
    - process_expert_data_for_factor_analysis_output.txt


Author: Nele Albers
Date: September 2023
"""

import pandas as pd


NUM_ACTIVITIES = 44  # Number of preparatory activities


def get_all_ratings_labels_exp(df, num_activities = 44):
    """
    Get all activity ratings and corresponding labels and explanations.

    Parameters
    ----------
    df : pandas DataFrame
        Qualtrics data from expert study.
    num_activities : int, optional
        Number of activities. The default is 44.

    Returns
    -------
    df_ratings : pandas DataFrame
        Activity ratings, labels, and explanations for all competencies.

    """

    # Titles of questions for labels in Qualtrics
    questions_labels = ["S1_Label1", "S1_Label2", 
                        "S2_Label1", "S2_Label2",
                        "PA1_Label1", "PA1_Label2", 
                        "PA2_Label1", "PA2_Label2"]
    # Titles of rating question stems in Qualtrics
    questions_rat_stems = ["_S1_Group1_R_1", "_S1_Group2_R_1", 
                           "_S2_Group1_R_1", "_S2_Group2_R_1",
                           "_PA1_Group1_R_1", "_PA1_Group2_R_1",
                           "_PA2_Group1_R_1", "_PA2_Group2_R_1"]
    # Full question titles for all activities
    questions_rat = [[str(i) + stem for i in range(1, num_activities + 1)] for stem in questions_rat_stems]

    all_ratings = []
    all_labels = []
    all_explanations = []
    
    for participant in range(len(df)):
        for label_idx, label in enumerate(questions_labels):
            ratings = df.iloc[participant][questions_rat[label_idx]].to_list()
            all_ratings.append(ratings)
            all_labels.append(df.iloc[participant][label])
            all_explanations.append(df.iloc[participant][label + "_Exp"])

    df_ratings = pd.DataFrame(all_ratings)
    df_ratings["Label"] = all_labels
    df_ratings["Explanation"] = all_explanations

    return df_ratings


if __name__ == "__main__":

    df = pd.read_excel("Expert_Competencies_Final_Qualtrics_Export_2022_12_22_anonym.xlsx")

    # Use complete responses only
    df = df[df["Finished"] == "True"]
    df = df.reset_index(drop=True)

    # Get the preparatory activity ratings and labels and explanations for all competencies the experts provided
    df_ratings = get_all_ratings_labels_exp(df, NUM_ACTIVITIES)

    # Transpose for use in factor analysis
    df_ratings.transpose().to_csv("all_expert_competency_ratings_transposed.csv", 
                                  index=False)

    # Let's also print the backgrounds and mean expertise ratings
    print("*** Self-Reported Expertise ***")
    print("Coaching for behavior change: M =", round(df["Expertise_1"].mean(), 2), "SD =", round(df["Expertise_1"].std(), 2))
    print("Coaching for quitting smoking: M =", round(df["Expertise_2"].mean(), 2), "SD =", round(df["Expertise_2"].std(), 2))
    print("Coaching for becoming more physically active: M =", round(df["Expertise_3"].mean(), 2), "SD =", round(df["Expertise_3"].std(), 2))

    print("\n*** Background ***")
    print(df["background"].to_list())

    # And let's also write this to a file
    with open("process_expert_data_for_factor_analysis_output.txt", 'w') as f:
        
        f.write("*** Self-Reported Expertise ***\n")
        f.write("Coaching for behavior change: M = " + str(round(df["Expertise_1"].mean(), 2)) + " SD = " + str(round(df["Expertise_1"].std(), 2)) + "\n")
        f.write("Coaching for quitting smoking: M = " +  str(round(df["Expertise_2"].mean(), 2)) + " SD = " + str(round(df["Expertise_2"].std(), 2)) + "\n")
        f.write("Coaching for becoming more physically active: M = " + str(round(df["Expertise_3"].mean(), 2)) + " SD = " + str(round(df["Expertise_3"].std(), 2)) + "\n")
        f.write("\n*** Background ***\n")
        f.write(str(df["background"].to_list()))
