# -*- coding: utf-8 -*-
"""
Get the activity ratings for competencies to be used
in the factor analysis for getting the views of users.

Required files:
    - user_responses_for_coding_final_full_dataframe_anonym.csv
    - user_responses_for_coding_final_labels_and_explanations_only_anonym_pa_coded.xlsx
    - user_responses_for_coding_final_labels_and_explanations_only_anonym_smoking_coded.xlsx
    
Output files:
    - all_competency_ratings_transposed.csv.
    - process_user_data_for_factor_analysis_output.txt


Author: Nele Albers
Date: September 2023
"""

import pandas as pd


def get_all_ratings(df_final, explanations_list, num_activities = 44):
    """
    Get all ratings corresponding to labels, based on the explanations for the
    labels.

    Parameters
    ----------
    df_final : pandas DataFrame
        Qualtrics data from user study.
    explanations_list : list
        Explanations we should consider.
    num_activities : int, optional
        Number of preparatory activities. The default is 44.

    Returns
    -------
    all_ratings : list
        All activity ratings for the explanations we should consider.

    """

    # Titles of questions for label explanations in Qualtrics
    questions_exp = ["S_Means_Label1_Exp", "S_Means_Label2_Exp",
                     "PA_Means_Label1_Exp", "PA_Means_Label2_Exp",
                     "S_Ends_Label1_Exp", "S_Ends_Label2_Exp",
                     "PA_Ends_Label1_Exp", "PA_Ends_Label2_Exp"]
    # Questions are called "Ends" but used for both means and ends in Qualtrics
    questions_rat_stems = ["_S_Ends_Group1_R_1", "_S_Ends_Group2_R_1",
                           "_PA_Ends_Group1_R_1", "_PA_Ends_Group2_R_1"]
    # Full question titles for all activities
    questions_rat = [[str(i) + stem for i in range(1, num_activities + 1)] for stem in questions_rat_stems]


    all_ratings = []
    for coded_idx, coded in enumerate(explanations_list):

        if coded_idx % 25 == 0:
            print("...", coded_idx)

        for q_idx, q in enumerate(questions_exp):

            # Get index of corresponding question stem
            stem_idx = q_idx
            if q_idx > 3:
                stem_idx -= 4

            for participant in range(len(df_final)):

                if coded == df_final.iloc[participant][q]:

                    ratings = df_final.iloc[participant][questions_rat[stem_idx]].to_list()
                    all_ratings.append(ratings)
        
    return all_ratings


if __name__ == "__main__":

    NUM_ACTIVITIES = 44  # Number of preparatory activities


    # Load coding by first coder
    df_code_pa = pd.read_excel("user_responses_for_coding_final_labels_and_explanations_only_anonym_pa_coded.xlsx")
    df_code_s = pd.read_excel("user_responses_for_coding_final_labels_and_explanations_only_anonym_smoking_coded.xlsx")
    df_code = pd.concat((df_code_pa, df_code_s))
    # Get responses for competencies
    df_code_c = df_code[df_code["Competency"] == 1]
    df_code_c = df_code_c.reset_index()
    num_labels_coded_comp = len(df_code_c)


    # Load final grouping questionnaire export
    df_final = pd.read_csv('user_responses_for_coding_final_full_dataframe_anonym.csv')


    # Get ratings for competencies
    print("... Getting ratings for competencies")
    all_comp_ratings = get_all_ratings(df_final, df_code_c["Explanation"].to_list())
    df_all_comp_ratings = pd.DataFrame(all_comp_ratings)
    df_all_comp_ratings["Label"] = df_code_c["Label"].to_list()
    df_all_comp_ratings["Explanation"] = df_code_c["Explanation"].to_list()
    df_all_comp_ratings.transpose().to_csv("all_competency_ratings_transposed.csv", 
                                           index=False)


    # Get number of participants providing competencies
    num_participants_comp = len(df_final[df_final["means_or_ends"] == "ends"])
    
    
    # Write to file
    with open("process_user_data_for_factor_analysis_output.txt", 'w') as f:
        f.write("Number of smoker participants providing competencies: " + str(num_participants_comp) + "\n")
        f.write("Number of labels coded as competencies: " + str(num_labels_coded_comp) + "\n")
        f.write("Number of labels coded as way of doing a preparatory activity: " + str(len(df_code[df_code["Means"] == 1])) + "\n")
        f.write("Number of labels coded as unclear: " + str(len(df_code[df_code["Unclear or Mixed"] == 1])) + "\n")
