# -*- coding: utf-8 -*-
"""
Compute each activity's contribution to each of the 6 expert-identified competencies.

Competencies:
    - Factor 1: Practical skills (+) - Clear future identity (-)
    - Factor 2: Motivation (+) - Knowledge (-)
    - Factor 3: Insights into personal strengths (+) - Insights into personal
                weaknesses (-)
                
Required files:
    - Data/comp_3factors_df_loadings_experts.csv
    - Data/comp_3factors_df_ratings_experts.csv

Output files:
    - activity_contribution_expert_competencies_raw.csv


Author: Nele Albers
Date: September 2023

"""

import numpy as np
import pandas as pd


if __name__ == "__main__":

    # Names of the competency columns
    COMPETENCY_COLS = ["Comp_0", "Comp_1", "Comp_2", "Comp_3", "Comp_4", "Comp_5"]

    # Ratings for the activities by experts
    df_ratings = pd.read_csv("Data/comp_3factors_df_ratings_experts.csv")
    df_ratings = df_ratings.drop(columns = ["Unnamed: 0"])

    # Weights from the factor analysis
    df_loadings = pd.read_csv("Data/comp_3factors_df_loadings_experts.csv")
    df_loadings = df_loadings.drop(columns = ["Unnamed: 0"])

    # Only use loadings of at least 0.4 magnitude
    for col in ["MR1", "MR2", "MR3"]:
        for i in range(len(df_loadings)):
            if abs(df_loadings.at[i, col]) < 0.4:
                df_loadings.at[i, col] = 0

    # Get the mapping of the activities onto the factors    
    df = np.matmul(df_ratings, df_loadings)

    for col_idx, col in enumerate([0, 1, 2]):
        df["Comp_" + str(col_idx * 2)] = [df.at[a, col]/df[col].max() if df.at[a, col] > 0 else 0 for a in range(len(df))]

    for col_idx, col in enumerate([0, 1, 2]):
        df["Comp_" + str(col_idx * 2 + 1)] = [df.at[a, col]/df[col].min() if df.at[a, col] < 0 else 0 for a in range(len(df))]

    # Order the competencies in ascending order
    df_raw = df[COMPETENCY_COLS]

    # Save to file
    df_raw.to_csv("activity_contribution_expert_competencies_raw.csv")
