# -*- coding: utf-8 -*-
"""
This file shows how we anonymized the data from the follow-up questionnaire.
Since we cannot provide the non-anonymized data, you cannot run this script yourself.

Required files:
    - anonymization_key.xlsx
    - qualtrics_followup_questionnaire_final_2023_10_26.csv
    
Created files:
    - clean_and_anonymize_followupquestionnaire_data_output.txt
    - followupquestionnaire_anonym.csv

Author: Nele Albers
Date: January 2024
"""

import pandas as pd


if __name__ == "__main__":

    # Load follow-up questionnaire data
    df_fu = pd.read_csv("qualtrics_followup_questionnaire_final_2023_10_26.csv")


    # Load anonymization key
    df_key = pd.read_excel("anonymization_key.xlsx")


    # Anonymize data
    df_fu = pd.merge(df_fu, df_key, on = "PROLIFIC_PID")
    df_fu = df_fu.drop(columns=["PROLIFIC_PID", "Prolific_ID"])


    # Remove the person who returned their completed submission on Prolific
    # Looked up this person's anonymized ID in the anonymization key file
    df_fu = df_fu[df_fu["rand_id"] != "P25"]


    # Remove columns we do not need/cannot publish
    df_fu = df_fu.drop(columns=["StartDate", "EndDate", "Status",
                                "Progress", "Duration (in seconds)",
                                "RecordedDate", "ResponseId",
                                "UserLanguage", "DistributionChannel",
                                "Q_RecaptchaScore", "Last_smoke",
                                "Godin_exercise_11", "Godin_exercise_9",
                                "Godin_exercise_8"])


    # Remove data from people who did not finish the questionnaire
    # One of these incomplete submissions is from a person who did not complete the
    # questionnaire in the first attempt but then completed it in the second attempt.
    # Gave same responses to the overlapping questions (weekly exercise amount),
    # so can use the complete submission.
    length_before = len(df_fu)
    df_fu = df_fu.dropna(subset=["Quitter_Self_Identit_2"])
    length_after = len(df_fu)
    print("Number of incomplete submissions removed (including one of a person who later completed):", length_before - length_after)
    with open("clean_and_anonymize_followupquestionnaire_data_output.txt", 'w') as f:
        f.write("Number of incomplete submissions removed (including one of a person who later completed): " + str(length_before - length_after) + "\n")


    # Remove data from people failing the attention check (there was only one)
    length_before = len(df_fu)
    df_fu = df_fu[df_fu["Quitter_Self_Identit_2"].isin(["Disagree strongly", "Disagree"])]
    df_fu = df_fu.reset_index(drop=True)
    length_after = len(df_fu)
    print("Number of submissions removed for failing AC:", length_before - length_after)
    print("Remaining number of submissions:", length_after)
    with open("clean_and_anonymize_followupquestionnaire_data_output.txt", 'a') as f:
        f.write("Number of submissions removed for failing attention check: " + str(length_before - length_after) + "\n")
        f.write("Remaining number of submissions: " + str(length_after))


    # Save anonymized data
    df_fu.to_csv("followupquestionnaire_anonym.csv", index=False)
