# -*- coding: utf-8 -*-
"""
Clean and anonymize the data from the prescreening questionnaire and Prolific profiles.
Since we cannot provide the non-anonymized data, you cannot run this script yourself.

We can only keep the data of people whom we paid for the prescreening. This are
people who (partially) completed session 1 and people who were not contemplating
or preparing to quit smoking.

Required files:
    - final_keep_track_of_participants.xlsx
    - Qualtrics_Exports/qualtrics_prescreening_14_2023_08_11.xlsx
    - Prolific profile data exports

Created files:
    - anonymization_key.xlsx
    - prescreening_questionnaire_anonym.csv
    - prolific_profile_anonym.csv

Author: Nele Albers
Date: January 2024
"""


import pandas as pd


if __name__ == "__main__":

    # Load our overview of the participants who took part in session 1
    df_s1 = pd.read_excel("final_keep_track_of_participants.xlsx", 'Session 1')
    df_took_s1 = df_s1[df_s1["Paid"] == 1]
    df_took_s1_ids = df_took_s1["ID"].to_list()
    
    # Qualtrics export of pre-screening questionnaire
    df_prescreening = pd.read_excel("Qualtrics_Exports/qualtrics_prescreening_14_2023_08_11.xlsx")
    
    
    # Keep only those people who (partially) completed session 1 or were not contemplating or
    # preparing to quit smoking
    df_prescreening = df_prescreening[(df_prescreening["PROLIFIC_PID"].isin(df_took_s1_ids)) | (df_prescreening["ttm_smoking"] == "No, NOT thinking of quitting")]
    # drop incomplete submissions for people who (partially) completed session 1
    df_prescreening = df_prescreening[(~pd.isna(df_prescreening["AC_2_1"])) | (df_prescreening["ttm_smoking"] == "No, NOT thinking of quitting")]
    df_prescreening = df_prescreening.reset_index(drop=True)
    
    
    # Merge data from Prolific profiles
    df_pre_1_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_1_2023_07_21.xlsx")
    df_pre_2_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_2_2023_07_23.xlsx")
    df_pre_3_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_3_2023_07_24.xlsx")
    df_pre_4_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_4_2023_07_25.xlsx")
    df_pre_5_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_5_2023_07_26.xlsx")
    df_pre_6_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_6_2023_07_27.xlsx")
    df_pre_7_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_7_2023_07_28.xlsx")
    df_pre_8_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_8_2023_07_31.xlsx")
    df_pre_9_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_9_2023_08_01.xlsx")
    df_pre_10_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_10_2023_08_03.xlsx")
    df_pre_11_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_11_2023_08_04.xlsx")
    df_pre_12_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_12_2023_08_07.xlsx")
    df_pre_13_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_13_2023_08_10.xlsx")
    df_pre_14_f = pd.read_excel("Prolific_Exports/prolific_export_prescreening_female_14_2023_08_11.xlsx")
    
    df_pre_1_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_1_2023_07_21.xlsx")
    df_pre_2_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_2_2023_07_23.xlsx")
    df_pre_3_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_3_2023_07_24.xlsx")
    df_pre_4_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_4_2023_07_25.xlsx")
    df_pre_5_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_5_2023_07_26.xlsx")
    df_pre_6_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_6_2023_07_27.xlsx")
    df_pre_7_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_7_2023_07_28.xlsx")
    df_pre_8_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_8_2023_07_31.xlsx")
    df_pre_9_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_9_2023_08_01.xlsx")
    df_pre_10_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_10_2023_08_03.xlsx")
    df_pre_11_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_11_2023_08_04.xlsx")
    df_pre_12_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_12_2023_08_07.xlsx")
    df_pre_13_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_13_2023_08_10.xlsx")
    df_pre_14_m = pd.read_excel("Prolific_Exports/prolific_export_prescreening_male_14_2023_08_11.xlsx")
    
    df_pre_prolific = pd.concat([df_pre_1_f, df_pre_2_f, df_pre_3_f, df_pre_4_f, df_pre_5_f,
                                 df_pre_1_m, df_pre_2_m, df_pre_3_m, df_pre_4_m, df_pre_5_m,
                                 df_pre_6_m, df_pre_6_f, df_pre_7_m, df_pre_7_f, df_pre_8_m,
                                 df_pre_8_f, df_pre_9_m, df_pre_9_f, df_pre_10_m, df_pre_10_f,
                                 df_pre_11_m, df_pre_11_f, df_pre_12_m, df_pre_12_f, df_pre_13_m,
                                 df_pre_13_f, df_pre_14_m, df_pre_14_f])
    
    
    # Keep only the Prolific profile data we can keep
    df_pre_prolific = df_pre_prolific[df_pre_prolific["Participant id"].isin(df_prescreening["PROLIFIC_PID"].to_list())]
    # Remove data from people with returned submissions
    df_pre_prolific = df_pre_prolific[df_pre_prolific["Age"]!= "CONSENT_REVOKED"]
    df_pre_prolific = df_pre_prolific.reset_index(drop=True)
    
    
    # Remove Qualtrics data from people with returned submissions on Prolific
    df_prescreening = df_prescreening[df_prescreening["PROLIFIC_PID"].isin(df_pre_prolific["Participant id"].to_list())]
    df_prescreening = df_prescreening.reset_index(drop=True)
    
    
    # Create random ID
    all_ids_sorted = sorted(df_prescreening["PROLIFIC_PID"].to_list())
    df_prescreening["rand_id"] = ["P" + str(all_ids_sorted.index(df_prescreening.at[i, "PROLIFIC_PID"]) + 1) for i in range(len(df_prescreening))]
    df_pre_prolific["rand_id"] = ["P" + str(all_ids_sorted.index(df_pre_prolific.at[i, "Participant id"]) + 1) for i in range(len(df_pre_prolific))]
    
    
    # Save anonymization key so we can later also anonymize the session data and post-questionnaire data
    # with the same random identifiers
    df_key = df_prescreening[["PROLIFIC_PID", "rand_id"]]
    df_key.to_excel("anonymization_key.xlsx", index=False)
    
    
    # Save anonymized pre-screening questionnaire data
    df_prescreening = df_prescreening.drop(columns = ["PROLIFIC_PID", "StartDate",
                                                      "EndDate", "Status", "Progress",
                                                      "Finished", "Duration (in seconds)",
                                                      "RecordedDate", "ResponseId",
                                                      "DistributionChannel", "UserLanguage",
                                                      "Q_RecaptchaScore", "Prolific_ID",
                                                      "PIV_English_Fluency"])
    # Also remove columns we do not need for this chapter
    df_prescreening = df_prescreening.drop(columns = ["Quit_Before_24_Hours",
                                                      "Num_Quit_24h_Last_Y",
                                                      "Time_Since_Last_Quit",
                                                      "Godin_exercise_11",
                                                      "Godin_exercise_8",
                                                      "Godin_exercise_9",
                                                      "Means_1_1", "Means_2_1",
                                                      "Means_3_1", "Means_4_1",
                                                      "Means_5_1"])
    df_prescreening.to_csv("prescreening_questionnaire_anonym.csv", index=False)
    
    
    # Save anonymized Prolific profile data
    df_pre_prolific = df_pre_prolific.drop(columns=["Participant id",
                                                    "Status"])
    # Also remove columns we do not need for this chapter
    df_pre_prolific = df_pre_prolific.drop(columns=["Weekly exercise",
                                                    "Smoking frequency"])
    df_pre_prolific.to_csv("prolific_profile_anonym.csv", index=False)
