'''
Feature selection and Q-value and dynamics computations for MDP.

Author: Nele Albers
Date: September 2023
'''


from copy import deepcopy
import itertools
import math
import numpy as np
import random
from scipy import stats


def approx_dynamics_blocks_independent_features(data, num_blocks_plus_1, num_act, blocks, 
                                                blocks_plus_1, f, feat_sel, 
                                                preparatory_actions_list,
                                                mean_impute_value,
                                                num_vals_per_feature = 2):
    """Approximate reward and transition functions for blocks. Transitions for the features thereby depend only on the single feature, not on other features.

    Args:
        data (list): List with samples of the form <s0, s1, a, r>.
        num_blocks_plus_1 (int): Number of blocks when using 1 more than the so far selected features.
        num_act (int): Number of actions.
        blocks (list): Abstract states based on so far selected features.
        blocks_plus_1 (list): Abstract states based on selecting one more feature.
        f (int): Candidate feature
        feat_sel (list): Already selected features
        preparatory_actions_list (list): List of preparatory activity cluster indices
        mean_impute_value (float): Value to use to fill in reward values if we have less than the desired minimum amount
                                   for a certain block.
        num_vals_per_feature (int): Number of values per feature
        
    Returns:
        np-array: reward function
        np-array: transition function
    """

    num_samples = len(data)
    min_count = 10  # min. number of samples we want per block for the reward function

    trans_func = np.zeros((num_blocks_plus_1, num_act, num_blocks_plus_1))
    # Reward function does not depend on the next state
    reward_func = np.zeros((num_blocks_plus_1, num_act))
    reward_func_count = np.zeros((num_blocks_plus_1, num_act))

    # Get reward function
    for b_ind, block in enumerate(blocks):  # for each block based on so far selected features

        for data_index in range(num_samples):  # for each data sample

            s = data[data_index][0][f]  # value for current feature 
            s_b = np.take(data[data_index][0], feat_sel)  # values for already selected features

            # the start state must be in the current block
            if list(s_b) == block:

                # get the action and reward
                a = data[data_index][2]
                r = data[data_index][3]

                # update statistics on rewards
                reward_func[blocks_plus_1.index(list(s_b) + [s]), a] += r
                reward_func_count[blocks_plus_1.index(list(s_b) + [s]), a] += 1

    # Now need to normalize the reward function
    # For each block including the current candidate feature
    for b_ind in range(num_blocks_plus_1):  
        for a in range(num_act):  # for each action
            if reward_func_count[b_ind, a] >= min_count:
                reward_func[b_ind, a] /= reward_func_count[b_ind, a]
            # We have a preparatory activity cluster with fewer samples than the minimum
            # So we fill in the data with the mean until we get min_count samples
            # and then we compute the overall average.
            elif a in preparatory_actions_list:
                reward_func[b_ind, a] += (min_count - reward_func_count[b_ind, a]) * mean_impute_value
                reward_func[b_ind, a] /= min_count
            # For non-preparatory activities we do not need the minimum number of samples because all rewards are the same.
            elif reward_func_count[b_ind, a] >= 1:
                reward_func[b_ind, a] /= reward_func_count[b_ind, a]

    # Now the transitions depend only on a single feature
    # So we compute for each single feature the transitions
    trans_func_single_list = []
    for feature in feat_sel + [f]:

        trans_func_single = np.zeros((num_vals_per_feature, num_act, num_vals_per_feature))

        for s in range(num_vals_per_feature):  # i.e. for different values of features
            for data_index in range(num_samples):  # for each data sample
                if data[data_index][0][feature] == s:
                    a = data[data_index][2]
                    s_prime = data[data_index][1][feature]  # feature value of next state
                    r = data[data_index][3]

                    # update statistics
                    trans_func_single[s, a, s_prime] += 1

            # Normalize transition function for this feature
            for a in range(num_act):
                summed = sum(trans_func_single[s, a])
                if summed > 0:
                    trans_func_single[s, a] /= summed
                # if we have no data on a feature value-action combination, we assume equal probability of transitioning to each other feature value
                else:
                    trans_func_single[s, a] = np.ones(num_vals_per_feature) / num_vals_per_feature  # i.e. equal probability of transitioning to each feature value

        trans_func_single_list.append(trans_func_single)


    # Now we need to compute the overall transition function
    # Afterward, we do not again need to normalize the transition function because these multiplied probabilities should add up to 1.
    for b_ind_1, block_1 in enumerate(blocks_plus_1):  # For each new block
        for a in range(num_act):  # For each action
            for b_ind_2, block_2 in enumerate(blocks_plus_1):  # For each next new block

                trans_func[b_ind_1, a, b_ind_2] = 1  # First initialize to 1 for the multiplication

                # For each feature in the new blocks, multiply their probabilities
                for f_idx in range(len(block_1)):
                    trans_func[b_ind_1, a, b_ind_2] *= trans_func_single_list[f_idx][block_1[f_idx], a, block_2[f_idx]]


    return reward_func, trans_func
        

def approx_dynamics_single_feature(data, num_act, f, num_vals_per_feature = 2):
    """Approximate reward and transition functions based on single features.

    Args:
        data (list): List with samples of the form <s0, s1, a, r>.
        num_act (int): Number of actions.
        f (int): Feature
        num_vals_per_feature (int): Number of values per feature
        
    Returns:
        np-array: reward function (num_vals_per_feature x num_actions)
        np-array: transition function (num_vals_per_feature x num_actions x num_vals_per_feature)
    """
    
    num_samples = len(data)
    
    trans_func = np.zeros((num_vals_per_feature, num_act, num_vals_per_feature))
    reward_func = np.zeros((num_vals_per_feature, num_act))
    reward_func_count = np.zeros((num_vals_per_feature, num_act))

    for s in range(num_vals_per_feature):  # i.e. for all possible feature values
        for data_index in range(num_samples):  # for each data sample
            if data[data_index][0][f] == s:
                a = data[data_index][2]
                s_prime = data[data_index][1][f]  # feature value of next state
                r = data[data_index][3]

                # update statistics
                trans_func[s, a, s_prime] += 1
                reward_func[s, a] += r
                reward_func_count[s, a] += 1

        # Normalize transition and reward function for current value of feature
        for a in range(num_act):
            summed = sum(trans_func[s, a])
            if summed > 0:
                trans_func[s, a] /= summed
            # if we have no data on a state-action combination, we assume equal probability of transitioning to each other state
            else:
                trans_func[s, a] = np.ones(num_vals_per_feature) / num_vals_per_feature  # i.e. equal probability of transitioning to each feature value
            if reward_func_count[s, a] > 0:
                reward_func[s, a] /= reward_func_count[s, a]

    return reward_func, trans_func


def feature_selection(data_in, effort_mean, feat_to_select, 
                      preparatory_actions_list,
                      num_feat_to_select = 3, 
                      num_act = 5,
                      discount_factor = 0.85):
    """Select state features in a way inspired by the G-algorithm based
    on Q-values.

    Args:
        data_in (list): List with samples of the form <s0, s1, a, r>.
        effort_mean (float): Mean of effort responses.
        feat_to_select (list): Candidate features.
        preparatory_actions_list (list): Indices of preparatory activity clusters.
        num_feat_to_select (int): Number of features to select.
        num_act (int): Number of possible actions.
        discount_factor (float): Discount factor for Q-value computation.

    Returns:
        list: Selected features.
        list: Criteria for selected features.
    """

    data = deepcopy(data_in)

    # All effort responses
    list_of_efforts = list(np.array(data, dtype=object)[:, 3].astype(float))

    # Map effort responses to rewards from -1 to 1, with the mean mapped to 0.
    map_to_rewards = get_map_effort_reward(effort_mean, 
                                           output_lower_bound = -1, 
                                           output_upper_bound = 1, 
                                           input_lower_bound = 0, 
                                           input_upper_bound = 10)
    reward_list = map_efforts_to_rewards(list_of_efforts, map_to_rewards)

    # now write these obtained reward values into "data" in place of the original rewards
    for i in range(len(reward_list)):
        data[i][3] = reward_list[i]

    num_feat = len(feat_to_select)

    # Select first feature
    q_values = np.zeros((num_feat, 2, num_act))
    t_tests = np.zeros((num_feat))

    for f in range(num_feat):

        # Compute transition function and reward function for feature
        reward_func, trans_func = approx_dynamics_single_feature(data, 
                                                                 num_act, f)

        # Value iteration for current feature    
        q_values[f], _ = get_Q_values_opt_policy(discount_factor, 
                                                 trans_func, 
                                                 reward_func, 
                                                 reward_dep_next_state = False)
        
        # t-test -> get p-value
        # one-sample t-test, comparing the absolute difference to 0
        t_tests[f] = stats.ttest_1samp(np.abs(np.array(q_values[f][0]) - np.array(q_values[f][1])), 0)[1]
    
    min_p_val = np.nanmin(t_tests)  # minimum p-value for t-test, ignoring nan-values
    feat_sel_options = [i for i in range(num_feat) if t_tests[i] == min_p_val]
    feat_sel = [random.choice(feat_sel_options)]  # choose randomly if there are multiple best features
    criterion = "First feature -> min. p-value: " + str(round(min_p_val, 4))
    if len(feat_sel_options) > 1:
        criterion += " random from " + str(feat_sel_options)
    feat_sel_criteria = [criterion]
    print("First feature selected:", feat_sel[0], criterion)

    # Select remaining features
    for j in range(num_feat_to_select - 1):
        num_feat_sel = len(feat_sel)
        feat_not_sel = [f for f in range(num_feat) if not f in feat_sel]
        num_feat_not_sel = len(feat_not_sel)

        # all abstract states based on the so far selected features
        blocks = [list(i) for i in itertools.product([0, 1], repeat = num_feat_sel)]
        # all abstract states including the new feature
        blocks_plus_1 = [list(i) for i in itertools.product([0, 1], repeat = num_feat_sel + 1)]
        num_blocks_plus_1 = len(blocks_plus_1)

        q_values_2 = np.zeros((num_feat_not_sel, num_blocks_plus_1, num_act))
        t_tests_2 = np.zeros(num_feat_not_sel)

        for f_ind, f in enumerate(feat_not_sel):  # for each not yet selected feature

            # Approximate reward and transition function
            reward_func, trans_func = approx_dynamics_blocks_independent_features(data, 
                                                                                  num_blocks_plus_1, 
                                                                                  num_act, 
                                                                                  blocks, 
                                                                                  blocks_plus_1, 
                                                                                  f,
                                                                                  feat_sel,
                                                                                  preparatory_actions_list = preparatory_actions_list,
                                                                                  mean_impute_value = 0)

            # Now we need to compute the Q-values
            # Value iteration for current feature    
            q_values_2[f_ind], _ = get_Q_values_opt_policy(discount_factor, 
                                                           trans_func, 
                                                           reward_func, 
                                                           reward_dep_next_state = False)

            # Separate Q-values based on whether current candidate feature is 0 or 1
            q_values_all_blocks_0 = np.array([q_values_2[f_ind][b_ind_even] for b_ind_even in range(0, len(blocks_plus_1), 2)]).flatten()
            q_values_all_blocks_1 = np.array([q_values_2[f_ind][b_ind_odd] for b_ind_odd in range(1, len(blocks_plus_1), 2)]).flatten()

            # Now that we have the data on all blocks, we run a one-sample t-test
            t_value = stats.ttest_1samp(np.abs(np.array(q_values_all_blocks_0) - np.array(q_values_all_blocks_1)), 0)[1]
            # If all differences are 0, the p-value will be nan -> so set the p-value high, i.e. 1
            if math.isnan(t_value):
                t_value = 1
            t_tests_2[f_ind] = t_value

        # Select next feature
        min_val_curr = np.min(t_tests_2)
        # Set tolerance for considering values equal to 10 ** -14
        feat_min_p_val = [i for i in range(num_feat_not_sel) if abs(t_tests_2[i] - min_val_curr) <= 10 ** -14]
        if len(feat_min_p_val) == 1:
            feat_sel.append(feat_not_sel[feat_min_p_val[0]])
            feat_sel_criteria.append("Min p-value: " + str(round(min_val_curr, 4)))
        # multiple best features -> choose one randomly
        else:
            feat_sel.append(feat_not_sel[random.choice(feat_min_p_val)])
            feat_sel_criteria.append("Min p-value: " + str(round(min_val_curr, 4)) + " random from " + str(feat_min_p_val))

        print("Feature selected:", feat_sel[-1])
        print("Criterion:", feat_sel_criteria[-1])

    return feat_sel, feat_sel_criteria


def compute_dynamics_effort(data_in, effort_mean, feat_sel, num_act, 
                            preparatory_actions_list,
                            comp_rewards = True):
    """Compute dynamics based on effort and its relevant state features only.

    Args:
        data_in (list): List with samples of the form <s0, s1, a, r>.
        effort_mean(float): Mean effort response to use for reward signal.
        feat_sel (list): Features to consider for policy computation.
        num_act (int): Number of possible actions.
        preparatory_actions_list (list): The indices of the preparatory actions.
        comp_rewards (boolean): Whether we need to compute rewards.

    Returns:
        np-array: reward function (num_states x num_actions)
        np-array: transition function (num_states x num_actions x num_states)

    """

    data = deepcopy(data_in)
    min_count = 10  # number of samples we want for each feature value combination for the rewards
    mean_mapped_effort = 0.5
    num_values_per_feature = 2

    # Map efforts to rewards
    if comp_rewards:
        # All effort responses
        list_of_efforts = list(np.array(data, dtype=object)[:, 3].astype(float))

        # Map effort responses to rewards from 0 to 1, with the mean mapped to 0.5.
        map_to_rewards = get_map_effort_reward(effort_mean, 
                                               output_lower_bound = 0, 
                                               output_upper_bound = 1, 
                                               input_lower_bound = 0, 
                                               input_upper_bound = 10)
        reward_list = map_efforts_to_rewards(list_of_efforts, map_to_rewards)
        # now write these obtained reward values into "data" in place of the original rewards
        for i in range(len(reward_list)):
            data[i][3] = reward_list[i]

    num_samples = len(data)
    num_feat_selected = len(feat_sel)

    abstract_states = [list(i) for i in itertools.product([0, 1], repeat = num_feat_selected)]

    # Compute reward function
    # Reward function is not dependent on next state
    reward_func = np.zeros((int(num_values_per_feature ** num_feat_selected), num_act))
    reward_func_count = np.zeros((int(num_values_per_feature ** num_feat_selected), num_act))

    for s_ind, s in enumerate(abstract_states):
        for data_index in range(num_samples):
            if list(np.take(np.array(data[data_index][0]), feat_sel)) == s:
                r = data[data_index][3]
                a = data[data_index][2]
                reward_func[s_ind, a] += r
                reward_func_count[s_ind, a] += 1

        # Normalize reward function
        for a in range(num_act):
            if reward_func_count[s_ind, a] >= min_count:
                reward_func[s_ind, a] /= reward_func_count[s_ind, a]
            # We have a preparatory activity cluster with fewer samples than the minimum
            # So we fill in the data with the mean until we get min_count samples
            # and then we compute the overall average.
            elif a in preparatory_actions_list:
                reward_func[s_ind, a] += (min_count - reward_func_count[s_ind, a]) * mean_mapped_effort
                reward_func[s_ind, a] /= min_count
            # For non-preparatory activities we do not need the minimum number of samples because all rewards are the same.
            elif reward_func_count[s_ind, a] >= 1:
                reward_func[s_ind, a] /= reward_func_count[s_ind, a]


    trans_func = np.zeros((int(num_values_per_feature ** num_feat_selected), num_act, int(num_values_per_feature ** num_feat_selected)))

    # Now the transitions depend only on a single feature
    # So we compute for each single feature the transitions
    trans_func_single_list = []
    for f in feat_sel:

        trans_func_single = np.zeros((num_values_per_feature, num_act, num_values_per_feature))

        for s in range(num_values_per_feature):  # i.e. if feature is 0 and if feature is 1
            for data_index in range(num_samples):  # for each data sample
                if data[data_index][0][f] == s:
                    a = data[data_index][2]
                    s_prime = data[data_index][1][f]  # feature value of next state

                    # update statistics
                    trans_func_single[s, a, s_prime] += 1

            # Normalize transition function for this feature
            for a in range(num_act):
                summed = sum(trans_func_single[s, a])
                if summed > 0:
                    trans_func_single[s, a] /= summed
                # if we have no data on a feature value-action combination, we assume equal probability of transitioning to each other feature value
                else:
                    trans_func_single[s, a] = np.ones(num_values_per_feature) / num_values_per_feature # i.e. equal probability of transitioning to each feature value

        trans_func_single_list.append(trans_func_single)

    # Now we need to compute the overall transition function
    # Afterward, we do not again need to normalize the transition function because these multiplied probabilities should add up to 1.
    for s_1_idx, s_1 in enumerate(abstract_states):  # For each state
        for a in range(num_act):  # For each action
            for s_2_idx, s_2 in enumerate(abstract_states):  # For each next state

                trans_func[s_1_idx, a, s_2_idx] = 1  # First initialize to 1 for the multiplication

                # For each feature in the state, multiply their probabilities
                for f_idx in range(len(s_1)):
                    trans_func[s_1_idx, a, s_2_idx] *= trans_func_single_list[f_idx][s_1[f_idx], a, s_2[f_idx]]

    return reward_func, trans_func


def compute_std_reward_effort(data_in, effort_mean, feat_sel, num_act,
                              preparatory_actions_list,
                              comp_rewards = True):
    """
    Compute the standard deviations for the effort-based reward.

    Parameters
    ----------
    data_in : list
        List with samples of the form <s0, s1, a, r>.
    effort_mean : float
        Mean effort response to use for reward signal.
    feat_sel : list
        User-inquired features to consider.
    num_act : int
        Number of actions.
    preparatory_actions_list : list
       Indices of preparatory actions
    comp_rewards : bool, optional
        Whether we need to compute rewards. The default is True.

    Returns
    -------
    reward_func_std : numpy array
        Standard deviation of effort-based rewards for each combination of 
        preparatory actions and user-inquired state feature values. Set to 0 for the 
        persuasive activities.

    """
    
    data = deepcopy(data_in)
    
    # Map efforts to rewards
    if comp_rewards:
        # All effort responses
        list_of_efforts = list(np.array(data, dtype=object)[:, 3].astype(int))

        # Map effort responses to rewards from 0 to 1, with the mean mapped to 0.5.
        map_to_rewards = get_map_effort_reward(effort_mean, 
                                               output_lower_bound = 0, 
                                               output_upper_bound = 1, 
                                               input_lower_bound = 0, 
                                               input_upper_bound = 10)
        reward_list = map_efforts_to_rewards(list_of_efforts, map_to_rewards)
        # now write these obtained reward values into "data" in place of the original rewards
        for i in range(len(reward_list)):
            data[i][3] = reward_list[i]
    
    num_samples = len(data)
    num_feat_selected = len(feat_sel)
    
    abstract_states = [list(i) for i in itertools.product([0, 1], repeat = num_feat_selected)]
    
    # Compute standard deviations for effort-based rewards
    reward_func_effort_std = np.zeros((int(2 ** num_feat_selected), num_act))
    
    for s_ind, s in enumerate(abstract_states):
        for a in range(len(preparatory_actions_list)):  # Only for the preparatory action clusters
            reward_func_s_a = []
            for data_index in range(num_samples):
                a_curr = data[data_index][2]
                if (list(np.take(np.array(data[data_index][0]), feat_sel)) == s) and a_curr == a:
                    reward_func_s_a.append(data[data_index][3])
            reward_func_effort_std[s_ind, a] = np.std(reward_func_s_a)
            
    return reward_func_effort_std


def get_map_effort_reward(effort_mean, output_lower_bound, 
                          output_upper_bound = 1, input_lower_bound = 0, 
                          input_upper_bound = 10):
    """
    Computes a mapping from effort responses to rewards.
    
    Args:
        effort_mean (float): mean effort response, to be mapped to halfway 
                             between output_lower_bound and output_upper_bound
        output_lower_bound (float): lowest value on output scale
        output_upper_bound (float, default 1): highest value on output scale
        input_lower_bound (int, default 0): lowest value on input scale, to 
                                            be mapped to output_lower_bound
        input_upper_bound (int, default 10): highest value on input scale, 
                                             to be mapped to output_upper_bound
                                             
    Returns:
        dictionary: maps effort responses to output scale values (float)
    """
    precision = 7
    map_to_rew = {}
    
    # We can already map the endpoints of the input scale to the output scale
    map_to_rew[input_lower_bound] = output_lower_bound
    map_to_rew[input_upper_bound] = output_upper_bound
    
    # The mean value on the output scale
    mean_output = (output_upper_bound - output_lower_bound)/2 + output_lower_bound
    output_length_half = mean_output - output_lower_bound
    
    input_length_lower_half = effort_mean - input_lower_bound
    input_length_upper_half = input_upper_bound - effort_mean
    
    inc_lower_half = output_length_half / input_length_lower_half
    inc_upper_half = output_length_half / input_length_upper_half
    
    # Compute output scale values for input scale below mean.
    idx = 1
    for i in range(input_lower_bound + 1, int(np.ceil(effort_mean))):
        map_to_rew[i] = output_lower_bound + idx * inc_lower_half
        idx += 1
    
    # Compute output scale values for input scale above mean.
    idx = 1
    for i in range(input_upper_bound - 1, int(np.floor(effort_mean)), -1):
        map_to_rew[i] = output_upper_bound - idx * inc_upper_half
        idx += 1
    
    # Also create mapping for the effort mean
    map_to_rew[round(effort_mean, precision)] = mean_output
    
    return map_to_rew


def map_efforts_to_rewards(list_of_efforts, map_to_rewards):
    """
    Maps effort responses to rewards.
    
    Args:
        list_of_efforts (list): effort responses to be mapped
        map_to_rewards (dictionary): maps effort responses to reward values (float)
    
    Returns:
        list (float): resulting reward values
    """
    rewards = []
    precision = 7
    for e in list_of_efforts:
        rewards.append(map_to_rewards[round(e, precision)])
    
    return rewards


def policy_evaluation(observation_space_size, discount_factor, trans_func,
                      reward_func, policy, q_vals, 
                      update_tolerance = 0.000000001,
                      reward_dep_next_state = False):
    """
    Returns the q-values for a specific policy.
    
    Args:
        observation_space_size: number of observations
        discount_factor: discount factor of MDP
        trans_func: transition function
        reward_func: reward function
        policy: policy to compute q-values for
        update_tolerance: precision
        reward_dep_next_state (boolean, default: False): whether reward depends also on the next state
    """
    q_vals_new = deepcopy(q_vals)
    num_act = len(trans_func[0])
    
    update = 1 # max. change in Q-value in one iteration
   
    if not reward_dep_next_state: # reward does not also depend on next state
        while update > update_tolerance:
            update = 0
            for s in range(observation_space_size):
                for a in range(num_act):
                    q_vals_new[s, a] = reward_func[s, a] + discount_factor * sum([trans_func[s, a, s_prime] * q_vals[s_prime, int(policy[s_prime])] for s_prime in range(observation_space_size)])
                    update = max(update, abs(q_vals[s, a] - q_vals_new[s, a]))
            q_vals = deepcopy(q_vals_new)
    
    else: # reward does also depend on the next state
        while update > update_tolerance:
            update = 0
            for s in range(observation_space_size):
               
                for a in range(num_act):
                    q_vals_new[s, a] = sum([trans_func[s, a, s_prime] * (discount_factor * q_vals[s_prime, int(policy[s_prime])] + reward_func[s, a, s_prime]) for s_prime in range(observation_space_size)])
                    
                    update = max(update, abs(q_vals[s, a] - q_vals_new[s, a]))
            q_vals = deepcopy(q_vals_new)
        
    return q_vals_new
    

def get_Q_values_opt_policy(discount_factor, trans_func, reward_func,
                            reward_dep_next_state = False):
    """
    Returns the Q-values for each state under the optimal policy as well
    as the optimal policy.
    
    Args:
        discount_factor: discount factor of MDP
        trans_func: transition function (dim.: num_states x num_actions x num_states)
        reward_func: reward function (dim.: num_states x num_actions)
        reward_dep_next_state (boolean, default: False): whether reward depends also on the next state
    """
    min_iterations = 100
    num_states = len(trans_func)
    num_act = len(trans_func[0])
    q_vals = np.zeros((num_states, num_act))

    policy = np.zeros(num_states)
    policy_new = np.ones(num_states)
    it = 0
    
    while (not np.array_equal(policy, policy_new)) or it < min_iterations:
        q_vals = policy_evaluation(num_states, discount_factor, 
                                   trans_func, reward_func, policy, q_vals,
                                   reward_dep_next_state = reward_dep_next_state)
        policy = policy_new
        policy_new = np.array([np.argmax(q_vals[s]) for s in range(num_states)])
        it += 1
    
    return q_vals, policy_new
