# -*- coding: utf-8 -*-
"""
Compute different optimal policies.

Required files:
    - Data/activity_contribution_expert_competencies_raw.csv
    - Data/data_rl_samples_binary.csv

Output files:
    - Intermediate_Results/
       - _qvals_4_6100_0.85
       - _qvals_equally_likely_4_6100_0.85
       - _reward_func_4_6100_0.85
       - _reward_func_effort_4_6100_0.85
       - _reward_func_no_user_states_4_6100_0.85
       - _reward_func_std_effort_4_6100_0.85
       - _trans_func_effort_4_6100_0.85
       - _V_4_6100_0.85
       - policy_4_6100_0.85.csv
       - policy_equally_likely_4_6100_0.85.csv
       - policy_no_user_states_4_6100_0.85.csv
       - policy_user_states_only_4_6100_0.85.csv
    

Author: Nele Albers
Date: May 2024
"""

import numpy as np
import pandas as pd
import pickle

import compute_dynamics_feat_sel as dyn
import optimal_policy_computations as pol_comp


PATH = "Intermediate_Results/" # pre-fix for path for storing results

NUM_ACTIONS = 14  # number of preparatory activity clusters + number of persuasive activities
action_clusters = [[18, 19, 20, 21],
                   [26, 27, 31, 32, 33, 34, 35],
                   [16, 17, 28, 29, 30, 37, 38, 39, 40, 41],
                   [2, 3, 6, 7, 8, 9, 12, 13, 36, 42, 43],
                   [0, 1, 4, 5, 10, 11, 14, 15, 22, 23, 24, 25]]
NUM_PREP_CLUSTERS = 5  # Number of preparatory activity clusters
NUM_ACTIVITIES = 53  # Total number of activities (preparatory + persuasive)

FEAT_SEL = [0, 6, 10]  # Indices of selected user-inquired features
NUM_FEAT_USER = len(FEAT_SEL)  # Number of user-inquired features
NUM_EXPERT_COMPETENCIES = 6
MAP_VALS = [0, 0.33, 0.67, 1]  # Possible values for expert competencies

DISCOUNT_FACTOR = 0.85


if __name__ == "__main__":

    ### Load transition data
    data = pd.read_csv("Data/data_rl_samples_binary.csv", 
                       converters={'s0': eval, 's1': eval})
    data_train = data.copy(deep=True)
    # Actions need to start at 0; right now they start at 1
    data_train["a"] = data_train["a"] - 1
    # Mean effort spent on preparatory activities
    effort_mean_prep = data_train[data_train['a'] <= 4]["effort"].mean()


    ### Combinations of values for user-inquired features
    states_users = [[i, j, k] for i in range(2) for j in range(2) for k in range(2)]
    num_states_users = len(states_users)
    
    
    path_to_save =  str(len(MAP_VALS)) + "_" + str(FEAT_SEL[0]) + str(FEAT_SEL[1]) + str(FEAT_SEL[2]) + "_" + str(DISCOUNT_FACTOR)
        
    
    ### Preparatory activity contributions to expert competencies
    exp_comp_contributions = pd.read_csv("Data/activity_contribution_expert_competencies_raw.csv")
    all_expert_competencies = ["Comp_0", "Comp_1", "Comp_2", "Comp_3", "Comp_4", "Comp_5"]


    ### Compute the reward and transitions for the user-inquired features only
    # We set the effort for persuasive activities to 0, because their reward will always be 0
    # because they do not contribute to the expert-identified competencies
    data_train.loc[data_train['a'] > NUM_PREP_CLUSTERS - 1, 'effort'] = 0

    # Effort-based reward is mapped to between 0 and 1
    # when we have less than 10 values for a feature value combination for a preparatory activity cluster,
    # we balance the reward with the mean.
    reward_func_effort, transitions_effort = dyn.compute_dynamics_effort(data_train[["s0", "s1", "a", "effort"]].values.tolist(), 
                                                                         effort_mean_prep, 
                                                                         FEAT_SEL, 
                                                                         NUM_ACTIONS, 
                                                                         preparatory_actions_list = [i for i in range(NUM_PREP_CLUSTERS)],
                                                                         comp_rewards = True)


    with open(PATH + "_reward_func_effort_" + path_to_save, "wb") as f:
        pickle.dump(reward_func_effort, f)
    with open(PATH + "_trans_func_effort_" + path_to_save, "wb") as f:
        pickle.dump(transitions_effort, f)

    # Standard deviations for the effort-based rewards
    reward_func_std_effort = dyn.compute_std_reward_effort(data_train[["s0", "s1", "a", "effort"]].values.tolist(), 
                                                           effort_mean_prep, 
                                                           FEAT_SEL, 
                                                           NUM_ACTIONS, 
                                                           preparatory_actions_list = [i for i in range(NUM_PREP_CLUSTERS)],
                                                           comp_rewards = True)

    with open(PATH + "_reward_func_std_effort_" + path_to_save, "wb") as f:
        pickle.dump(reward_func_std_effort, f)


    ### Optimal policy based on the entire model
    optimal_policy_full, transition_func, reward_func, V = pol_comp.get_optimal_policy_full(FEAT_SEL, MAP_VALS, 
                                                                                            action_clusters, 
                                                                                            reward_func_effort, 
                                                                                            transitions_effort, 
                                                                                            exp_comp_contributions, 
                                                                                            all_expert_competencies, 
                                                                                            DISCOUNT_FACTOR,
                                                                                            num_vals_per_user_feature = 2,
                                                                                            log = False,
                                                                                            log_factor = 1)

    # Save the computed policy as well as the reward function and value function.
    # We need the transition functions for the simulations in our analysis, for example.
    df_policy = pd.DataFrame(optimal_policy_full)

    df_policy.to_csv(PATH + "policy_" + path_to_save + ".csv", index=False)

    with open(PATH + "_reward_func_" + path_to_save, "wb") as f:
        pickle.dump(reward_func, f)

    with open(PATH + "_V_" + path_to_save, 'wb') as f:
        pickle.dump(V, f)


    # Compute Q-values based on the optimal policy, using the value function V
    num_states = len(V)
    q_vals = np.zeros((num_states, NUM_ACTIVITIES))
    for s_idx in range(num_states):
        for a in range(NUM_ACTIVITIES):
            q_vals[s_idx, a] = reward_func[s_idx, a] + DISCOUNT_FACTOR * sum(transition_func[a][s_idx, :] * V)

    with open(PATH + "_qvals_" + path_to_save, 'wb') as f:
        pickle.dump(q_vals, f)


    ### Equally likely next values for user-inquired feature values
    # Transition function for equally likely next user states
    transitions_equally_likely_effort = np.ones((len(states_users), NUM_ACTIONS, len(states_users))) * 1/len(states_users)
    # Compute optimal policy for this case
    optimal_policy_equally_likely_next_user_states, trans_func_equally_likely, reward_func, V_equally_likely= pol_comp.get_optimal_policy_full(FEAT_SEL, MAP_VALS, 
                                                                                                                                               action_clusters, 
                                                                                                                                               reward_func_effort, 
                                                                                                                                               transitions_equally_likely_effort, 
                                                                                                                                               exp_comp_contributions, 
                                                                                                                                               all_expert_competencies, 
                                                                                                                                               DISCOUNT_FACTOR,
                                                                                                                                               num_vals_per_user_feature = 2,
                                                                                                                                               log = False,
                                                                                                                                               log_factor = 1)

    df_policy_equally_likely_next_user_states = pd.DataFrame(optimal_policy_equally_likely_next_user_states)
    df_policy_equally_likely_next_user_states.to_csv(PATH + "policy_equally_likely_" + path_to_save + ".csv", 
                                                     index=False)
    
    # Compute Q-values based on equally likely next user states, using the value function V
    num_states = len(V_equally_likely)
    q_vals_equally_likely = np.zeros((num_states, NUM_ACTIVITIES))
    for s_idx in range(num_states):
        for a in range(NUM_ACTIVITIES):
            q_vals_equally_likely[s_idx, a] = reward_func[s_idx, a] + DISCOUNT_FACTOR * sum(trans_func_equally_likely[a][s_idx, :] * V_equally_likely)

    with open(PATH + "_qvals_equally_likely_" + path_to_save, 'wb') as f:
        pickle.dump(q_vals_equally_likely, f)


    ### Optimal policy when considering no user states (i.e., we assume that people always spend the average effort)
    optimal_policy_no_user_states, reward_func_no_user_states = pol_comp.get_optimal_policy_only_expert_competency_states(FEAT_SEL, 
                                                                                                                          MAP_VALS, 
                                                                                                                          action_clusters, 
                                                                                                                          exp_comp_contributions, 
                                                                                                                          all_expert_competencies,
                                                                                                                          effort_mean = 0.5,
                                                                                                                          num_vals_per_user_feature = 2,
                                                                                                                          log = False,
                                                                                                                          log_factor = 1)

    df_policy_no_user_states = pd.DataFrame(optimal_policy_no_user_states)
    df_policy_no_user_states.to_csv(PATH + "policy_no_user_states_" + path_to_save + ".csv", 
                                    index=False)
    
    with open(PATH + "_reward_func_no_user_states_" + path_to_save, 'wb') as f:
        pickle.dump(reward_func_no_user_states, f)
    
    
    ### Optimal policy when considering the effort only 
    # (i.e., just pick the activity cluster with highest effort)
    optimal_policy_user_states_only = pol_comp.get_optimal_policy_only_user_states(FEAT_SEL, MAP_VALS, 
                                                                                   action_clusters,
                                                                                   reward_func_effort,
                                                                                   num_vals_per_user_feature = 2)

    df_policy_user_states_only = pd.DataFrame(optimal_policy_user_states_only)
    df_policy_user_states_only.to_csv(PATH + "policy_user_states_only_" + path_to_save + ".csv", 
                                      index=False)
