# -*- coding: utf-8 -*-
"""
Functions for optimal policy computations.

Author: Nele Albers
Date: September 2023.
"""

import mdptoolbox
import numpy as np
from scipy import sparse


NUM_ACTIVITIES = 53  # total number of activities (preparatory + persuasive)
NUM_PREPARATORY_ACTIVITIES = 44
NUM_PREP_CLUSTERS = 5  # number of clusters of preparatory activities


def get_optimal_policy_full(feat_sel, map_vals, action_clusters,
                            reward_func_effort, transitions_effort,
                            exp_comp_contributions,
                            all_expert_competencies, 
                            discount_factor,
                            num_vals_per_user_feature,
                            log = False, log_factor = 1):
    """
    Computes the full optimal policy.

    Parameters
    ----------
    feat_sel : list
        Indices of selected user features
    map_vals : list
        Possible values for expert competencies
    action_clusters : list
        Lists with indices of activities in clusters of preparatory activities
    reward_func_effort : np.array
        Reward function when considering user states and effort only.
    transitions_effort : np.array
        Transition function when considering user states only.
    exp_comp_contributions : pd.DataFrame
        Contributions of preparatory activities to expert competencies
    all_expert_competencies : list
        Names of expert competencies in dataframe
    discount_factor : float
        Discount factor for optimal policy computation
    num_vals_per_user_feature : int
        Number of values per user feature
    log : bool, optional
        Whether to use the log for the reward computation. The default is False.
    log_factor : float, optional
        Factor to use for log for reward computation. The default is 1.

    Returns
    -------
    list
        Optimal policy
    sparse matrix
        Transition function
    np.array
        Reward function
    tuple
        V* per state
    """
    
    STATES = [[i, j, k, l, m, n, o, p, q] for i in range(num_vals_per_user_feature) for j in range(num_vals_per_user_feature) for k in range(num_vals_per_user_feature) for l in map_vals for m in map_vals for n in map_vals for o in map_vals for p in map_vals for q in map_vals]
    STATES_USERS = [[i, j, k] for i in range(num_vals_per_user_feature) for j in range(num_vals_per_user_feature) for k in range(num_vals_per_user_feature)]
    NUM_STATES = len(STATES)
    
    NUM_USER_FEAT = len(feat_sel)
    NUM_EXPERT_COMPETENCIES = len(all_expert_competencies)
    
    reward_func = np.zeros((NUM_STATES, NUM_ACTIVITIES))

    transition_func = []
    for activity in range(NUM_ACTIVITIES):
        transition_func.append(sparse.lil_matrix((NUM_STATES, NUM_STATES),
                               dtype="float64"))
        
    for s_idx in range(NUM_STATES):
        if s_idx % 1000 == 0:
            print(s_idx)
        # First the transitions for the preparatory activities only
        for action_cluster_idx, action_cluster in enumerate(action_clusters):
            # get the effort for this cluster of preparatory activities
            effort = reward_func_effort[STATES_USERS.index(STATES[s_idx][0:NUM_USER_FEAT]), action_cluster_idx]
            for action in action_cluster:
                potential_contributions = np.array([exp_comp_contributions.at[action, comp] for comp in all_expert_competencies])
                # multiply potential contribution by effort and add to current expert-competency level
                effort_contributions = potential_contributions * effort + np.array(STATES[s_idx][NUM_USER_FEAT:])
                # map all resulting values to the possible expert competency values to get the next state
                next_state_experts = [map_vals[min(range(len(map_vals)), key = lambda i: abs(map_vals[i]-effort_contributions[c]))] for c in range(NUM_EXPERT_COMPETENCIES)]
                # now we can compute the overall reward as the actual increase in the expert competencies
                # so our reward is between 0 and 6 (in practice, there is no activity that contributes to all expert competencies with 1)
                if not log:
                    reward = sum(abs(np.array(next_state_experts) - np.array(STATES[s_idx][NUM_USER_FEAT:])))
                # here we take the natural logarithm of the current and next expert competency value after adding 1
                # this means that adding 0.5 to a current level of 0 is more valuable than adding 0.5 to a current level of 0.5, for example
                else:
                    reward = sum([np.log(next_state_experts[e_idx] * log_factor + 1) - np.log(STATES[s_idx][NUM_USER_FEAT + e_idx] * log_factor + 1) for e_idx in range(NUM_EXPERT_COMPETENCIES)])
                # the new value for the expert competencies is fixed, so we just need
                # to iterate through the different user states for the transitions
                for s_2_users in STATES_USERS:
                    s_2_idx = STATES.index(s_2_users + next_state_experts)
                    transition_func[action][s_idx, s_2_idx] = transitions_effort[STATES_USERS.index(STATES[s_idx][0:NUM_USER_FEAT]), action_cluster_idx, STATES_USERS.index(s_2_users)]
                # set reward, does not depend on next state
                reward_func[s_idx, action] = reward
        
        # Now for the persuasive activities
        for pers_activity in range(NUM_PREPARATORY_ACTIVITIES, NUM_ACTIVITIES):
            # persuasive activities do not change the expert competencies, so the expert competencies always stay the same when taking a persuasive activity
            # so we only need to iterate through the different user states
            for s_2_users in STATES_USERS:
                s_2_idx = STATES.index(s_2_users + STATES[s_idx][NUM_USER_FEAT:])
                transition_func[pers_activity][ s_idx, s_2_idx] = transitions_effort[STATES_USERS.index(STATES[s_idx][0:NUM_USER_FEAT]), pers_activity - NUM_ACTIVITIES + NUM_PREP_CLUSTERS, STATES_USERS.index(s_2_users)]
                # reward for persuasive activities is always 0
                # do not need to do anything, since initialization is 0


    for a in range(NUM_ACTIVITIES):
        transition_func[a] = transition_func[a].tocsr()

    m = mdptoolbox.mdp.ValueIterationGS(transition_func, reward_func, 
                                        discount = discount_factor, 
                                        epsilon=0.001, max_iter=1000)
    m.setVerbose()
    m.run()
    
    return m.policy, transition_func, reward_func, m.V


def get_optimal_policy_only_expert_competency_states(feat_sel, map_vals, action_clusters,
                                                     exp_comp_contributions,
                                                     all_expert_competencies,
                                                     effort_mean,
                                                     num_vals_per_user_feature,
                                                     log = False, log_factor = 1):
    """
    Compute the optimal policy when basically only considering expert competency states
    (i.e., we assume that people always spend the average effort).

    Parameters
    ----------
    feat_sel : list
        Indices of selected user features
    map_vals : list
        Possible values for expert competencies
    action_clusters : list
        Lists with indices of activities in clusters of preparatory activities
    exp_comp_contributions : pd.DataFrame
        Contributions of preparatory activities to expert competencies
    all_expert_competencies : list
        Names of expert competencies in dataframe.
    effort_mean: float
        Mean effort for preparatory activities.
    num_vals_per_user_feature : int
        Number of values per user feature
    log : bool, optional
        Whether to use the log for the reward computation. The default is False.
    log_factor : float, optional
        Factor to use for log for reward computation. The default is 1.

    Returns
    -------
    list
       Optimal policy for each state.
    numpy array
       Underlying reward function.

    """
    
    STATES = [[i, j, k, l, m, n, o, p, q] for i in range(num_vals_per_user_feature) for j in range(num_vals_per_user_feature) for k in range(num_vals_per_user_feature) for l in map_vals for m in map_vals for n in map_vals for o in map_vals for p in map_vals for q in map_vals]
    NUM_STATES = len(STATES)
    
    NUM_USER_FEAT = len(feat_sel)
    NUM_EXPERT_COMPETENCIES = len(all_expert_competencies)
    
    reward_func = np.zeros((NUM_STATES, NUM_ACTIVITIES))
        
    for s_idx in range(NUM_STATES):
        if s_idx % 1000 == 0:
            print(s_idx)
        # First the transitions for the preparatory activities only
        for action_cluster_idx, action_cluster in enumerate(action_clusters):
            for action in action_cluster:
                potential_contributions = np.array([exp_comp_contributions.at[action, comp] for comp in all_expert_competencies])
                # multiply potential contribution by mean effort and add to current expert-competency level
                effort_contributions = potential_contributions * effort_mean + np.array(STATES[s_idx][NUM_USER_FEAT:])
                # map all resulting values to the possible expert competency values to get the next state
                next_state_experts = [map_vals[min(range(len(map_vals)), key = lambda i: abs(map_vals[i]-effort_contributions[c]))] for c in range(NUM_EXPERT_COMPETENCIES)]
                # now we can compute the overall reward as the actual increase in the expert competencies
                # so our reward is between 0 and 6 (in practice, there is no activity that contributes to all expert competencies with 1)
                if not log:
                    reward = sum(abs(np.array(next_state_experts) - np.array(STATES[s_idx][NUM_USER_FEAT:])))
                # here we take the natural logarithm of the current and next expert competency value after adding 1
                # this means that adding 0.5 to a current level of 0 is more valuable than adding 0.5 to a current level of 0.5, for example
                else:
                    reward = sum([np.log(next_state_experts[e_idx] * log_factor + 1) - np.log(STATES[s_idx][NUM_USER_FEAT + e_idx] * log_factor + 1) for e_idx in range(NUM_EXPERT_COMPETENCIES)])
                reward_func[s_idx, action] = reward
    
    return np.argmax(reward_func, axis = 1), reward_func


def get_optimal_policy_only_user_states(feat_sel, map_vals, action_clusters,
                                        reward_func_effort,
                                        num_vals_per_user_feature):
    """
    Compute the optimal policy when only considering user states.
    So just calculate the preparatory activity cluster with highest effort.

    Parameters
    ----------
    feat_sel : list
        Indices of selected user features
    map_vals : list
        Possible values for expert competencies
    action_clusters : list
        Lists with indices of activities in clusters of preparatory activities
    reward_func_effort : np.array
        Reward function when considering user states and effort only.
    num_vals_per_user_feature : int
        Number of values per user feature.

    Returns
    -------
    list
       Optimal preparatory activity cluster for each state.

    """
    
    STATES = [[i, j, k, l, m, n, o, p, q] for i in range(num_vals_per_user_feature) for j in range(num_vals_per_user_feature) for k in range(num_vals_per_user_feature) for l in map_vals for m in map_vals for n in map_vals for o in map_vals for p in map_vals for q in map_vals]
    STATES_USERS = [[i, j, k] for i in range(num_vals_per_user_feature) for j in range(num_vals_per_user_feature) for k in range(num_vals_per_user_feature)]
    NUM_STATES = len(STATES)
    
    NUM_USER_FEAT = len(feat_sel)
    
    reward_func = np.zeros((NUM_STATES, NUM_PREP_CLUSTERS))
        
    for s_idx in range(NUM_STATES):
        if s_idx % 1000 == 0:
            print(s_idx)
        # First the transitions for the preparatory activities only
        for action_cluster_idx, action_cluster in enumerate(action_clusters):
            # Get the effort for this cluster of preparatory activities
            effort = reward_func_effort[STATES_USERS.index(STATES[s_idx][0:NUM_USER_FEAT]), action_cluster_idx]
            reward_func[s_idx, action_cluster_idx] = effort
    
    # Return preparatory activity cluster per state
    return np.argmax(reward_func, axis = 1)
