# -*- coding: utf-8 -*-
"""
Merge data from the sessions, Prolific profiles, and the pre-screening
questionnaire and create file with transition samples.

Required files:
    - Data/prescreening_questionnaire_anonym.csv
    - Data/prolific_profile_anonym.csv
    - Data/sessionsdata_anonym.csv
    
Output files:
    - AQ2/Data/data_rl_samples_binary.csv
    - DataOverview_ParticipantCharacteristics_FeatureSelection_AQ1/Data/data_rl_samples_binary.csv
    - merge_data_output.txt
    

Author: Nele Albers
Date: May 2024
"""

from copy import deepcopy
import numpy as np
import pandas as pd


if __name__ == "__main__":
    
    with open("merge_data_output.txt", 'w') as f:
        f.write("***Merging data***\n")


    NUM_SESSIONS = 5
    STATE_FEATURE_NAMES = ["state_1", "state_2", "state_3", "state_4", "state_5",
                           "state_6", "state_7", "state_8", "state_9", "state_busy",
                           "state_energy"]
    NUM_STATE_FEATURES = len(STATE_FEATURE_NAMES)


    ### Load data
    df_session = pd.read_csv("Data/sessionsdata_anonym.csv")
    df_prolific = pd.read_csv("Data/prolific_profile_anonym.csv")
    df_prolific = df_prolific[["Age", "rand_id", "Highest education level completed",
                               "Gender"]]
    df_prescreening = pd.read_csv("Data/prescreening_questionnaire_anonym.csv")
    df_prescreening = df_prescreening[["rand_id", "PIV_Smoking_Freq", "ttm_pa",
                                       "Weekly_Exercise"]]


    ### Collect transition samples
    all_first_states  = []
    all_next_states = []
    session_nums = []
    all_efforts = []
    all_actions = []
    all_ids = []
    all_activities = []
    all_states = []  # all states to compute means
    all_dropouts = []
    

    for session in range(1, NUM_SESSIONS):


        # All IDs of people with at least an answer to state_1 for the next session
        all_session_ids = list(set(df_session[(df_session["session_num"] == session + 1) & (df_session["response_type"] == "state_1")]["rand_id"].to_list()))


        for i in all_session_ids:

            # Get data from this person
            df_id = df_session[df_session["rand_id"] == i]


            # Get the first state
            state = []
            for state_feature_name in STATE_FEATURE_NAMES:
                state.append(int(df_id[(df_id["session_num"] == session) & (df_id["response_type"] == state_feature_name)]["response_value"]))


            # Get the next state
            next_state = []
            df_id_session_states = df_id[(df_id["session_num"] == session + 1) & (df_id["response_type"].isin(STATE_FEATURE_NAMES))]
            for state_feature_name in STATE_FEATURE_NAMES:
                if state_feature_name in df_id_session_states["response_type"].to_list():
                    next_state.append(float(df_id_session_states[df_id_session_states["response_type"] == state_feature_name]["response_value"]))
            # Fill up with nan if we do not have data on all state features
            num_features_obtained = len(next_state)
            for f in range(len(next_state), NUM_STATE_FEATURES):
                if f == num_features_obtained:
                    print("Next state with missing feature values. Only have data on " + str(len(next_state)) + " feature(s).")
                    with open("merge_data_output.txt", 'a') as f:
                        f.write("Next state with missing feature values. Only have data on " + str(len(next_state)) + " feature(s).\n")
                next_state.append(np.nan)


            all_first_states.append(state)
            all_next_states.append(next_state)
            session_nums.append(session)
            all_efforts.append(int(df_id[(df_id["session_num"] == session + 1) & (df_id["response_type"] == "effort")]["response_value"]))
            all_dropouts.append(int(df_id[(df_id["session_num"] == session + 1) & (df_id["response_type"] == "dropout_response")]["response_value"]))
            all_actions.append(int(df_id[(df_id["session_num"] == session) & (df_id["response_type"] == "cluster_new_index")]["response_value"]))
            all_activities.append(int(df_id[(df_id["session_num"] == session) & (df_id["response_type"] == "activity_new_index")]["response_value"]))
            all_ids.append(i)

            
            # Add the state to the list of all states for computing feature means
            all_states.append(deepcopy(state))


            # To get all states, we also need to add the next state if that next state is the last state of the person
            # By counting the number of times the person filled in information for the state_1 state feature,
            # we know how many sessions that person completed (almost) fully (some state features might be missing
            # from the last state).
            df_id_state_importance = df_id[df_id["response_type"] == "state_1"]
            if session == len(df_id_state_importance) - 1:
                all_states.append(deepcopy(next_state))
                
    
    ### To get all states, we also need to add the first states of people who do not have data for at least the first state feature in session 2
    # Get all IDs of people with data for at least the first state feature in session 2
    all_ids_arrived_in_session2 = list(set(df_session[(df_session["session_num"] == 2) & (df_session["response_type"] == "state_1")]["rand_id"].to_list()))
    # IDs of people with data for at least the first state feature in session 1 who do not have data for at least the first state feature in session 2
    all_ids_arrived_in_session1 = list(set(df_session[(df_session["session_num"] == 1) & (df_session["response_type"] == "state_1")]["rand_id"].to_list()))
    all_ids_arrived_in_session1_but_not_session2 = [i for i in all_ids_arrived_in_session1 if not i in all_ids_arrived_in_session2]
    print("Number of people with data on first state only:", len(all_ids_arrived_in_session1_but_not_session2))
    with open("merge_data_output.txt", 'a') as f:
        f.write("Number of people with data on first state only: " + str(len(all_ids_arrived_in_session1_but_not_session2)) + "\n")

    
    # Get the first states of those people with no first state feature data in session 2
    for i in all_ids_arrived_in_session1_but_not_session2:

        # Get data from this person
        df_id = df_session[df_session["rand_id"] == i]


        # Get the first state
        first_state = []
        df_id_session_states = df_id[(df_id["session_num"] == 1) & (df_id["response_type"].isin(STATE_FEATURE_NAMES))]
        for state_feature_name in STATE_FEATURE_NAMES:
            if state_feature_name in df_id_session_states["response_type"].to_list():
                first_state.append(float(df_id_session_states[df_id_session_states["response_type"] == state_feature_name]["response_value"]))
        # Fill up with nan if we do not have data on all state features
        num_features_obtained = len(first_state)
        for f in range(len(first_state), NUM_STATE_FEATURES):
            if f == num_features_obtained:
                print("First state with missing feature values. Only have data on " + str(len(first_state)) + " feature(s).")
                with open("merge_data_output.txt", 'a') as f:
                    f.write("First state with missing feature values. Only have data on " + str(len(first_state)) + " feature(s).\n")
            first_state.append(np.nan)


        all_states.append(first_state)


    ### Create dataframe with RL samples
    df_rl_samples = pd.DataFrame()
    df_rl_samples["rand_id"] = all_ids
    df_rl_samples["session"] = session_nums
    df_rl_samples["s0"] = all_first_states
    df_rl_samples["a"] = all_actions
    df_rl_samples["effort"] = all_efforts
    df_rl_samples["s1"] = all_next_states 
    df_rl_samples["activity"] = all_activities
    df_rl_samples["dropout_response"] = all_dropouts


    ### Now we also want to add the characteristics of the people corresponding to the samples
    df_rl_samples = pd.merge(df_rl_samples, df_prolific, on = "rand_id")
    df_rl_samples = pd.merge(df_rl_samples, df_prescreening, on = "rand_id")


    ### And we also want to create separate columns with the state features
    df_rl_samples[['s0_B0', "s0_B1", "s0_B2", "s0_B3", "s0_B4", "s0_B5", "s0_B6", "s0_B7", "s0_B8", "s0_Available_Time", "s0_Energy"]] = pd.DataFrame(df_rl_samples.s0.tolist(), index= df_rl_samples.index)
    df_rl_samples[['s1_B0', "s1_B1", "s1_B2", "s1_B3", "s1_B4", "s1_B5", "s1_B6", "s1_B7", "s1_B8", "s1_Available_Time", "s1_Energy"]] = pd.DataFrame(df_rl_samples.s1.tolist(), index= df_rl_samples.index)


    ### Compute binary feature values
    df_rl_samples_binary = deepcopy(df_rl_samples)


    # Compute feature means
    feat_means = np.nanmean(all_states, axis = 0)
    print("Feature means:", np.round(feat_means, 2))
    with open("merge_data_output.txt", 'a') as f:
        f.write("Feature means: " + str(np.round(feat_means, 2)))


    # We compute a binary version of the state features based on the feature means.
    # If we have partial data on a state for a session, we set the remaining missing
    # values equal to 1 (i.e., equal to the feature mean)
    for row in range(len(df_rl_samples_binary)):
            df_rl_samples_binary.at[row, 's0'] = [0 if df_rl_samples_binary.loc[row, 's0'][i] < feat_means[i] else 1 for i in range(NUM_STATE_FEATURES)]
            df_rl_samples_binary.at[row, 's1'] = [0 if df_rl_samples_binary.loc[row, 's1'][i] < feat_means[i] else 1 for i in range(NUM_STATE_FEATURES)]


    ### Save data
    df_rl_samples_binary.to_csv("DataOverview_ParticipantCharacteristics_FeatureSelection_AQ1/Data/data_rl_samples_binary.csv", 
                                index=False)
    df_rl_samples_binary.to_csv("AQ2/Data/data_rl_samples_binary.csv", 
                                index=False)
