# -*- coding: utf-8 -*-
"""
Preprocess data for computing changes between the prescreening questionnaire
and post-questionnaire and follow-up questionnaire
w.r.t. weekly exercise, smoking frequency, and quitter
self-identity. Also compute Cronbach's  alpha for the three items for quitter
self-identity.

Required files:
    - Data/followupquestionnaire_anonym.csv
    - Data/postquestionnaire_anonym.csv
    - Data/prescreening_questionnaire_anonym.csv
    
Created files:
    - Analysis_Time_Comparison/Data/prefu_preprocessed.csv
    - Analysis_Time_Comparison/Data/prepost_preprocessed.csv
    - process_data_for_time_comparison_output.txt

Author: Nele Albers
Date: May 2024
"""

import numpy as np
import pandas as pd


def cronbach_alpha(df):
    """
    Calculate cronbach's alpha.
    Code from https://towardsdatascience.com/cronbachs-alpha-theory-and-application-in-python-d2915dd63586.
    """

    # 1. Transform the df into a correlation matrix
    df_corr = df.corr()

    # 2.1 Calculate N
    # The number of variables equals the number of columns in the df
    N = df.shape[1]

    # 2.2 Calculate R
    # For this, we'll loop through the columns and append every
    # relevant correlation to an array called "r_s". Then, we'll
    # calculate the mean of "r_s"
    rs = np.array([])
    for i, col in enumerate(df_corr.columns):
        sum_ = df_corr[col][i+1:].values
        rs = np.append(sum_, rs)
    mean_r = np.mean(rs)

    # 3. Use the formula to calculate Cronbach's Alpha 
    cronbach_alpha = (N * mean_r) / (1 + (N - 1) * mean_r)

    return cronbach_alpha


def process_changes(df_1, df_2, df_changes_filename = "prepost_preprocessed"):
    """
    Create preprocessed dataframes for analyses on the changes between pre- and 
    post-measurements.

    Parameters
    ----------
    df_1 : pandas dataframe
        Pre-measurements
    df_2 : pandas dataframe
        Post-measurements
    df_changes_filename : string, optional
        Name for the csv-file with preprocessed data on changes. 
        The default is "prepost_preprocessed".

    Returns
    -------
    None.

    """

    df = pd.merge(df_1, df_2, on = "rand_id")


    ### Format and save preprocessed data for changes in pre- and post-measurements
    df_we = df[["rand_id", "Weekly_Exercise_y", "Weekly_Exercise_x"]].melt("rand_id")
    df_we["variable"] = df_we["variable"].map({'Weekly_Exercise_y': 2, 'Weekly_Exercise_x': 1})
    df_we = df_we.rename({'variable':"pre_post", 'value':"weekly_exercise"}, axis=1)
    
    df_qsi = df[["rand_id", "QSI_y", "QSI_x"]].melt("rand_id")
    df_qsi["variable"] = df_qsi["variable"].map({'QSI_y': 2, 'QSI_x': 1})
    df_qsi = df_qsi.rename({'variable':"pre_post", 'value':'QSI'}, axis=1)
    
    df_sf = df[["rand_id", "Smoking_Freq", "PIV_Smoking_Freq"]].melt("rand_id")
    df_sf["variable"] = df_sf["variable"].map({'Smoking_Freq': 2, 'PIV_Smoking_Freq': 1})
    df_sf = df_sf.rename({'variable':"pre_post", 'value':"smoking_frequency"}, axis=1)


    ### Merge sub-dataframes
    df_change = pd.merge(df_we, df_qsi, on = ['rand_id', 'pre_post'])
    df_change = pd.merge(df_change, df_sf, on = ['rand_id', "pre_post"])


    ### Add participant ID for multi-level model
    all_ids = df_change["rand_id"].to_list()
    df_change["pid"] = [all_ids.index(df_change.iloc[i]["rand_id"]) + 1 for i in range(len(df_change))]


    df_change.to_csv(df_changes_filename + ".csv", index=False)


if __name__ == "__main__":

    ### Load the three types of data
    df_pre = pd.read_csv("Data/prescreening_questionnaire_anonym.csv")
    df_post = pd.read_csv("Data/postquestionnaire_anonym.csv")
    df_fu = pd.read_csv("Data/followupquestionnaire_anonym.csv")


    ### Use only prescreening data on people who completed the post-questionnaire
    df_pre = df_pre[df_pre["rand_id"].isin(df_post["rand_id"].to_list())]


    ### Make relevant columns numeric
    df_pre["Weekly_Exercise"] = df_pre["Weekly_Exercise"].map({'Sometimes (60 – 150 minutes per week)': 1, 'Often (more than 150 minutes per week)': 2,
                                                               'Never (0 – 60 minutes per week)': 0})
    df_post["Weekly_Exercise"] = df_post["Weekly_Exercise"].map({'Sometimes (60 – 150 minutes per week)': 1, 'Often (more than 150 minutes per week)': 2,
                                                                 'Never (0 – 60 minutes per week)': 0})
    df_fu["Weekly_Exercise"] = df_fu["Weekly_Exercise"].map({'Sometimes (60 – 150 minutes per week)': 1, 'Often (more than 150 minutes per week)': 2,
                                                             'Never (0 – 60 minutes per week)': 0})

    df_pre["PIV_Smoking_Freq"] = df_pre["PIV_Smoking_Freq"].map({'Once a day': 0, '2-5 times a day': 1,
                                                                 '6-10 times a day': 2, '11-19 times a day': 3, 'More than 20 times a day': 4})
    df_post["Smoking_Freq"] = df_post["Smoking_Freq"].map({'Once a day': 0, '2-5 times a day': 1,
                                                           '6-10 times a day': 2, '11-19 times a day': 3, 'More than 20 times a day': 4,
                                                           "1-6 times a week": -1,
                                                           "Less than 4 times a month": -2, "Never": -3})
    df_fu["Smoking_Freq"] = df_fu["Smoking_Freq"].map({'Once a day': 0, '2-5 times a day': 1,
                                                       '6-10 times a day': 2, '11-19 times a day': 3, 'More than 20 times a day': 4,
                                                       "1-6 times a week": -1,
                                                       "Less than 4 times a month": -2, "Never": -3})

    # The third and fifth quitter self-identity items are attention check questions in the prescreening questionnaire
    for col in ["Quitter_SelfIdentity_1", "Quitter_SelfIdentity_2", "Quitter_SelfIdentity_4"]:
        df_pre[col] = df_pre[col].map({"Disagree strongly": 0, "Disagree": 1, "Neither agree nor disagree": 2,
                                       "Agree": 3, "Agree strongly": 4})
    for col in ["Quitter_Self_Identit_1", "Quitter_Self_Identit_2", "Quitter_Self_Identit_3"]:
        df_post[col] = df_post[col].map({"Disagree strongly": 0, "Disagree": 1, "Neither agree nor disagree": 2,
                                         "Agree": 3, "Agree strongly": 4})
    # The second quitter self-identity item is an attention check in the follow-up questionnaire
    for col in ["Quitter_Self_Identit_1", "Quitter_Self_Identit_3", "Quitter_Self_Identit_4"]:
        df_fu[col] = df_fu[col].map({"Disagree strongly": 0, "Disagree": 1, "Neither agree nor disagree": 2,
                                     "Agree": 3, "Agree strongly": 4})


    ### Compute mean quitter self-identity based on corresponding three items
    # First compute Cronbach's alpha
    # Note that we compute Cronbach's alpha only based on the data used for the time 
    # comparison analysis and not all data that we have for the prescreening questionnaire
    # (i.e., only based on the prescreening questionnaire data of the people who
    # also completed the post-questionnaire).
    ca_pre = cronbach_alpha(df_pre[["Quitter_SelfIdentity_1", "Quitter_SelfIdentity_2", "Quitter_SelfIdentity_4"]])
    ca_post = cronbach_alpha(df_post[["Quitter_Self_Identit_1", "Quitter_Self_Identit_2", "Quitter_Self_Identit_3"]])
    ca_fu = cronbach_alpha(df_fu[["Quitter_Self_Identit_1", "Quitter_Self_Identit_3", "Quitter_Self_Identit_4"]])
    with open("process_data_for_time_comparison_output.txt", 'w') as f:
        f.write("Cronbach's alpha quitter self-identity prescreening questionnaire: " + str(np.round(ca_pre, 2)) + " (N = " + str(len(df_pre)) + ")\n")
        f.write("Cronbach's alpha quitter self-identity post-questionnaire: " + str(np.round(ca_post, 2)) + " (N = " + str(len(df_post)) + ")\n")
        f.write("Cronbach's alpha quitter self-identity follow-up questionnaire: " + str(np.round(ca_fu, 2)) + " (N = " + str(len(df_fu)) + ")")

    # Compute means
    df_pre["QSI"] = (df_pre["Quitter_SelfIdentity_1"] + df_pre["Quitter_SelfIdentity_2"] + df_pre["Quitter_SelfIdentity_4"])/3
    df_post["QSI"] = (df_post["Quitter_Self_Identit_1"] + df_post["Quitter_Self_Identit_2"] + df_post["Quitter_Self_Identit_3"])/3
    df_fu["QSI"] = (df_fu["Quitter_Self_Identit_1"] + df_fu["Quitter_Self_Identit_3"] + df_fu["Quitter_Self_Identit_4"])/3


    ### Create preprocessed data for prescreening questionnaire vs.
    # post-questionnaire
    process_changes(df_pre, df_post, df_changes_filename = "Analysis_Time_Comparison/Data/prepost_preprocessed")


    ### Create preprocessed data for prescreening questionnaire vs.
    # follow-up questionnaire
    # Use only prescreening data for people who completed the follow-up
    df_pre = df_pre[df_pre["rand_id"].isin(df_fu["rand_id"].to_list())]
    process_changes(df_pre, df_fu, df_changes_filename = "Analysis_Time_Comparison/Data/prefu_preprocessed")
